import enum
import inspect
import sys
from abc import ABC, abstractmethod  # pylint: disable=[no-name-in-module]
from typing import Any, Callable, Dict, Generic, List, Optional, Set, Tuple, Type, TypeVar, Union, cast

import yaml

from knot_resolver.utils.functional import all_matches

from .base_generic_type_wrapper import BaseGenericTypeWrapper
from .base_value_type import BaseValueType
from .exceptions import AggregateDataValidationError, DataDescriptionError, DataValidationError
from .renaming import Renamed, renamed
from .types import (
    get_annotations,
    get_generic_type_argument,
    get_generic_type_arguments,
    get_generic_type_wrapper_argument,
    get_optional_inner_type,
    is_dict,
    is_enum,
    is_generic_type_wrapper,
    is_internal_field_name,
    is_list,
    is_literal,
    is_none_type,
    is_optional,
    is_tuple,
    is_union,
)

T = TypeVar("T")


def is_obj_type(obj: Any, types: Union[type, Tuple[Any, ...], Tuple[type, ...]]) -> bool:
    # To check specific type we are using 'type()' instead of 'isinstance()'
    # because for example 'bool' is instance of 'int', 'isinstance(False, int)' returns True.
    # pylint: disable=unidiomatic-typecheck
    if isinstance(types, tuple):
        return type(obj) in types
    return type(obj) is types


class Serializable(ABC):
    """An interface for making classes serializable to a dictionary (and in turn into a JSON)."""

    @abstractmethod
    def to_dict(self) -> Dict[Any, Any]:
        raise NotImplementedError(f"...for class {self.__class__.__name__}")

    @staticmethod
    def is_serializable(typ: Type[Any]) -> bool:
        return (
            typ in {str, bool, int, float}
            or is_none_type(typ)
            or is_literal(typ)
            or is_dict(typ)
            or is_list(typ)
            or is_generic_type_wrapper(typ)
            or (inspect.isclass(typ) and issubclass(typ, Serializable))
            or (inspect.isclass(typ) and issubclass(typ, BaseValueType))
            or (inspect.isclass(typ) and issubclass(typ, BaseSchema))
            or (is_optional(typ) and Serializable.is_serializable(get_optional_inner_type(typ)))
            or (is_union(typ) and all_matches(Serializable.is_serializable, get_generic_type_arguments(typ)))
        )

    @staticmethod
    def serialize(obj: Any) -> Any:
        if isinstance(obj, Serializable):
            return obj.to_dict()

        if isinstance(obj, (BaseValueType, BaseGenericTypeWrapper)):
            o = obj.serialize()
            # if Serializable.is_serializable(o):
            return Serializable.serialize(o)
            # return o

        if isinstance(obj, list):
            res: List[Any] = [Serializable.serialize(i) for i in cast(List[Any], obj)]
            return res

        return obj


class _LazyDefault(Generic[T], Serializable):
    """
    Wrapper for default values BaseSchema classes.

    Defers the default instantiation until the schema
    itself is being instantiated
    """

    def __init__(self, constructor: Callable[..., T], *args: Any, **kwargs: Any) -> None:
        # pylint: disable=[super-init-not-called]
        self._func = constructor
        self._args = args
        self._kwargs = kwargs

    def instantiate(self) -> T:
        return self._func(*self._args, **self._kwargs)

    def to_dict(self) -> Dict[Any, Any]:
        return Serializable.serialize(self.instantiate())


def lazy_default(constructor: Callable[..., T], *args: Any, **kwargs: Any) -> T:
    """We use a factory function because you can't lie about the return type in `__new__`."""
    return _LazyDefault(constructor, *args, **kwargs)  # type: ignore[return-value]


def _split_docstring(docstring: str) -> Tuple[str, Optional[str]]:
    """Split docstring into description of the class and description of attributes."""
    if "---" not in docstring:
        return ("\n".join([s.strip() for s in docstring.splitlines()]).strip(), None)

    doc, attrs_doc = docstring.split("---", maxsplit=1)
    return (
        "\n".join([s.strip() for s in doc.splitlines()]).strip(),
        attrs_doc,
    )


def _parse_attrs_docstrings(docstring: str) -> Optional[Dict[str, str]]:
    """Given a docstring of a BaseSchema, return a dict with descriptions of individual attributes."""
    _, attrs_doc = _split_docstring(docstring)
    if attrs_doc is None:
        return None

    # try to parse it as yaml:
    data = yaml.safe_load(attrs_doc)
    assert isinstance(data, dict), "Invalid format of attribute description"
    return cast(Dict[str, str], data)


def _get_properties_schema(typ: Type[Any]) -> Dict[Any, Any]:
    schema: Dict[Any, Any] = {}
    annot = get_annotations(typ)

    docstring: str = typ.__dict__.get("__doc__", "") or ""
    attribute_documentation = _parse_attrs_docstrings(docstring)
    for field_name, python_type in annot.items():
        name = field_name.replace("_", "-")
        schema[name] = _describe_type(python_type)

        # description
        if attribute_documentation is not None:
            if field_name not in attribute_documentation:
                raise DataDescriptionError(f"The docstring does not describe field '{field_name}'", str(typ))
            schema[name]["description"] = attribute_documentation[field_name]
            del attribute_documentation[field_name]

        # default value
        if hasattr(typ, field_name):
            assert Serializable.is_serializable(
                python_type
            ), f"Type '{python_type}' does not appear to be JSON serializable"
            schema[name]["default"] = Serializable.serialize(getattr(typ, field_name))

    if attribute_documentation is not None and len(attribute_documentation) > 0:
        raise DataDescriptionError(
            f"The docstring describes attributes which are not present - {tuple(attribute_documentation.keys())}",
            str(typ),
        )

    return schema


def _describe_type(typ: Type[Any]) -> Dict[Any, Any]:  # noqa: C901, PLR0911, PLR0912
    # pylint: disable=too-many-branches

    if inspect.isclass(typ) and issubclass(typ, BaseSchema):
        return typ.json_schema(include_schema_definition=False)

    if inspect.isclass(typ) and issubclass(typ, BaseValueType):
        return typ.json_schema()

    if is_generic_type_wrapper(typ):
        wrapped = get_generic_type_wrapper_argument(typ)
        return _describe_type(wrapped)

    if is_none_type(typ):
        return {"type": "null"}

    if typ is int:
        return {"type": "integer"}

    if typ is bool:
        return {"type": "boolean"}

    if typ is str:
        return {"type": "string"}

    if is_literal(typ):
        lit: List[Any] = []
        args = get_generic_type_arguments(typ)
        for arg in args:
            if is_literal(arg):
                lit += get_generic_type_arguments(arg)
            else:
                lit.append(arg)
        return {"type": "string", "enum": lit}

    if is_optional(typ):
        desc = _describe_type(get_optional_inner_type(typ))
        if "type" in desc:
            desc["type"] = [desc["type"], "null"]
            return desc
        return {"anyOf": [{"type": "null"}, desc]}

    if is_union(typ):
        variants = get_generic_type_arguments(typ)
        return {"anyOf": [_describe_type(v) for v in variants]}

    if is_list(typ):
        return {"type": "array", "items": _describe_type(get_generic_type_argument(typ))}

    if is_dict(typ):
        key, val = get_generic_type_arguments(typ)

        if inspect.isclass(key) and issubclass(key, BaseValueType):
            assert (
                key.__str__ is not BaseValueType.__str__
            ), "To support derived 'BaseValueType', __str__ must be implemented."
        else:
            assert key is str, "We currently do not support any other keys then strings"

        return {"type": "object", "additionalProperties": _describe_type(val)}

    if inspect.isclass(typ) and issubclass(typ, enum.Enum):  # same as our is_enum(typ), but inlined for type checker
        return {"type": "string", "enum": [str(v) for v in typ]}

    raise NotImplementedError(f"Trying to get JSON schema for type '{typ}', which is not implemented")


TSource = Union[None, "BaseSchema", Dict[str, Any]]


def _create_untouchable(name: str) -> object:
    class _Untouchable:
        def __getattribute__(self, item_name: str) -> Any:
            raise RuntimeError(f"You are not supposed to access object '{name}'.")

        def __setattr__(self, item_name: str, value: Any) -> None:
            raise RuntimeError(f"You are not supposed to access object '{name}'.")

    return _Untouchable()


class ObjectMapper:
    def _create_tuple(self, tp: Type[Any], obj: Tuple[Any, ...], object_path: str) -> Tuple[Any, ...]:
        types = get_generic_type_arguments(tp)
        errs: List[DataValidationError] = []
        res: List[Any] = []
        for i, (t, val) in enumerate(zip(types, obj)):
            try:
                res.append(self.map_object(t, val, object_path=f"{object_path}[{i}]"))
            except DataValidationError as e:
                errs.append(e)
        if len(errs) == 1:
            raise errs[0]
        if len(errs) > 1:
            raise AggregateDataValidationError(object_path, child_exceptions=errs)
        return tuple(res)

    def _create_dict(self, tp: Type[Any], obj: Dict[Any, Any], object_path: str) -> Dict[Any, Any]:
        key_type, val_type = get_generic_type_arguments(tp)
        try:
            errs: List[DataValidationError] = []
            res: Dict[Any, Any] = {}
            for key, val in obj.items():
                try:
                    nkey = self.map_object(key_type, key, object_path=f"{object_path}[{key}]")
                    nval = self.map_object(val_type, val, object_path=f"{object_path}[{key}]")
                    res[nkey] = nval
                except DataValidationError as e:
                    errs.append(e)
            if len(errs) == 1:
                raise errs[0]
            if len(errs) > 1:
                raise AggregateDataValidationError(object_path, child_exceptions=errs)
        except AttributeError as e:
            raise DataValidationError(
                f"Expected dict-like object, but failed to access its .items() method. Value was {obj}", object_path
            ) from e
        else:
            return res

    def _create_list(self, tp: Type[Any], obj: List[Any], object_path: str) -> List[Any]:
        if isinstance(obj, str):
            raise DataValidationError("expected list, got string", object_path)

        inner_type = get_generic_type_argument(tp)
        errs: List[DataValidationError] = []
        res: List[Any] = []

        try:
            for i, val in enumerate(obj):
                res.append(self.map_object(inner_type, val, object_path=f"{object_path}[{i}]"))
            if len(res) == 0:
                raise DataValidationError("empty list is not allowed", object_path)
        except DataValidationError as e:
            errs.append(e)
        except TypeError as e:
            errs.append(DataValidationError(str(e), object_path))

        if len(errs) == 1:
            raise errs[0]
        if len(errs) > 1:
            raise AggregateDataValidationError(object_path, child_exceptions=errs)
        return res

    def _create_str(self, obj: Any, object_path: str) -> str:
        # we are willing to cast any primitive value to string, but no compound values are allowed
        if is_obj_type(obj, (str, float, int)) or isinstance(obj, BaseValueType):
            return str(obj)
        if is_obj_type(obj, bool):
            raise DataValidationError(
                "Expected str, found bool. Be careful, that YAML parsers consider even"
                ' "no" and "yes" as a bool. Search for the Norway Problem for more'
                " details. And please use quotes explicitly.",
                object_path,
            )
        raise DataValidationError(
            f"expected str (or number that would be cast to string), but found type {type(obj)}", object_path
        )

    def _create_int(self, obj: Any, object_path: str) -> int:
        # we don't want to make an int out of anything else than other int
        # except for BaseValueType class instances
        if is_obj_type(obj, int) or isinstance(obj, BaseValueType):
            return int(obj)
        raise DataValidationError(f"expected int, found {type(obj)}", object_path)

    def _create_union(self, tp: Type[T], obj: Any, object_path: str) -> T:
        variants = get_generic_type_arguments(tp)
        errs: List[DataValidationError] = []
        for v in variants:
            try:
                return self.map_object(v, obj, object_path=object_path)
            except DataValidationError as e:
                errs.append(e)

        raise DataValidationError("could not parse any of the possible variants", object_path, child_exceptions=errs)

    def _create_optional(self, tp: Type[Optional[T]], obj: Any, object_path: str) -> Optional[T]:
        inner: Type[Any] = get_optional_inner_type(tp)
        if obj is None:
            return None
        return self.map_object(inner, obj, object_path=object_path)

    def _create_bool(self, obj: Any, object_path: str) -> bool:
        if is_obj_type(obj, bool):
            return obj
        raise DataValidationError(f"expected bool, found {type(obj)}", object_path)

    def _create_literal(self, tp: Type[Any], obj: Any, object_path: str) -> Any:
        args = get_generic_type_arguments(tp)

        expected = []
        if sys.version_info < (3, 9):
            for arg in args:
                if is_literal(arg):
                    expected += get_generic_type_arguments(arg)
                else:
                    expected.append(arg)
        else:
            expected = args

        if obj in expected:
            return obj
        raise DataValidationError(f"'{obj}' does not match any of the expected values {expected}", object_path)

    def _create_base_schema_object(self, tp: Type[Any], obj: Any, object_path: str) -> "BaseSchema":
        if isinstance(obj, (dict, BaseSchema)):
            return tp(obj, object_path=object_path)
        raise DataValidationError(f"expected 'dict' or 'NoRenameBaseSchema' object, found '{type(obj)}'", object_path)

    def create_value_type_object(self, tp: Type[Any], obj: Any, object_path: str) -> "BaseValueType":
        if isinstance(obj, tp):
            # if we already have a custom value type, just pass it through
            return obj
        # no validation performed, the implementation does it in the constuctor
        try:
            return tp(obj, object_path=object_path)
        except ValueError as e:
            if len(e.args) > 0 and isinstance(e.args[0], str):
                msg = e.args[0]
            else:
                msg = f"Failed to validate value against {tp} type"
            raise DataValidationError(msg, object_path) from e

    def _create_default(self, obj: Any) -> Any:
        if isinstance(obj, _LazyDefault):
            return obj.instantiate()
        return obj

    def map_object(  # noqa: C901, PLR0911, PLR0912
        self,
        tp: Type[Any],
        obj: Any,
        default: Any = ...,
        use_default: bool = False,
        object_path: str = "/",
    ) -> Any:
        """
        Given an expected type `cls` and a value object `obj`.

        Return a new object of the given type and map fields of `obj` into it.
        During the mapping procedure, runtime type checking is performed.
        """
        # Disabling these checks, because I think it's much more readable as a single function
        # and it's not that large at this point. If it got larger, then we should definitely split it
        # pylint: disable=too-many-branches,too-many-locals,too-many-statements

        # default values
        if obj is None and use_default:
            return self._create_default(default)

        # NoneType
        if is_none_type(tp):
            if obj is None:
                return None
            raise DataValidationError(f"expected None, found '{obj}'.", object_path)

        # Optional[T]  (could be technically handled by Union[*variants], but this way we have better error reporting)
        if is_optional(tp):
            return self._create_optional(tp, obj, object_path)

        # Union[*variants]
        if is_union(tp):
            return self._create_union(tp, obj, object_path)

        # after this, there is no place for a None object
        if obj is None:
            raise DataValidationError(f"unexpected value 'None' for type {tp}", object_path)

        # int
        if tp is int:
            return self._create_int(obj, object_path)

        # str
        if tp is str:
            return self._create_str(obj, object_path)

        # bool
        if tp is bool:
            return self._create_bool(obj, object_path)

        # float
        if tp is float:
            raise NotImplementedError(
                "Floating point values are not supported in the object mapper."
                " Please implement them and be careful with type coercions"
            )

        # Literal[T]
        if is_literal(tp):
            return self._create_literal(tp, obj, object_path)

        # Dict[K,V]
        if is_dict(tp):
            return self._create_dict(tp, obj, object_path)

        # any Enums (probably used only internally in DataValidator)
        if is_enum(tp):
            if isinstance(obj, tp):
                return obj
            raise DataValidationError(f"unexpected value '{obj}' for enum '{tp}'", object_path)

        # List[T]
        if is_list(tp):
            return self._create_list(tp, obj, object_path)

        # Tuple[A,B,C,D,...]
        if is_tuple(tp):
            return self._create_tuple(tp, obj, object_path)

        # type of obj and cls type match
        if is_obj_type(obj, tp):
            return obj

        # when the specified type is Any, just return the given value
        # on mypy version 1.11.0 comparison-overlap error started popping up
        # https://github.com/python/mypy/issues/17665
        if tp == Any:  # type: ignore[comparison-overlap]
            return obj

        # BaseValueType subclasses
        if inspect.isclass(tp) and issubclass(tp, BaseValueType):
            return self.create_value_type_object(tp, obj, object_path)

        # BaseGenericTypeWrapper subclasses
        if is_generic_type_wrapper(tp):
            inner_type = get_generic_type_wrapper_argument(tp)
            obj_valid = self.map_object(inner_type, obj, object_path)
            return tp(obj_valid, object_path=object_path)

        # nested BaseSchema subclasses
        if inspect.isclass(tp) and issubclass(tp, BaseSchema):
            return self._create_base_schema_object(tp, obj, object_path)

        # if the object matches, just pass it through
        if inspect.isclass(tp) and isinstance(obj, tp):
            return obj

        # default error handler
        raise DataValidationError(
            f"Type {tp} cannot be parsed. This is a implementation error. "
            "Please fix your types in the class or improve the parser/validator.",
            object_path,
        )

    def is_obj_type_valid(self, obj: Any, tp: Type[Any]) -> bool:
        """Runtime type checking. Validate, that a given object is of a given type."""
        try:
            self.map_object(tp, obj)
        except (DataValidationError, ValueError):
            return False
        else:
            return True

    def _assign_default(self, obj: Any, name: str, python_type: Any, object_path: str) -> None:
        cls = obj.__class__

        try:
            default = self._create_default(getattr(cls, name, None))
        except ValueError as e:
            raise DataValidationError(str(e), f"{object_path}/{name}") from e

        value = self.map_object(python_type, default, object_path=f"{object_path}/{name}")
        setattr(obj, name, value)

    def _assign_field(self, obj: Any, name: str, python_type: Any, value: Any, object_path: str) -> None:
        value = self.map_object(python_type, value, object_path=f"{object_path}/{name}")
        setattr(obj, name, value)

    def _assign_fields(self, obj: Any, source: Union[Dict[str, Any], "BaseSchema", None], object_path: str) -> Set[str]:  # noqa: C901
        """
        Assign fields and values.

        Order of assignment:
          1. all direct assignments
          2. assignments with conversion method
        """
        cls = obj.__class__
        annot = get_annotations(cls)
        errs: List[DataValidationError] = []

        used_keys: Set[str] = set()
        for name, python_type in annot.items():
            try:
                if is_internal_field_name(name):
                    continue

                # populate field
                if source is None:
                    self._assign_default(obj, name, python_type, object_path)

                # check for invalid configuration with both transformation function and default value
                elif hasattr(obj, f"_{name}") and hasattr(obj, name):
                    raise RuntimeError(
                        f"Field '{obj.__class__.__name__}.{name}' has default value and transformation function at"
                        " the same time. That is now allowed. Store the default in the transformation function."
                    )

                # there is a transformation function to create the value
                elif hasattr(obj, f"_{name}") and callable(getattr(obj, f"_{name}")):
                    val = self._get_converted_value(obj, name, source, object_path)
                    self._assign_field(obj, name, python_type, val, object_path)
                    used_keys.add(name)

                # source just contains the value
                elif name in source:
                    val = source[name]
                    self._assign_field(obj, name, python_type, val, object_path)
                    used_keys.add(name)

                # there is a default value, or the type is optional => store the default or null
                elif hasattr(obj, name) or is_optional(python_type):
                    self._assign_default(obj, name, python_type, object_path)

                # we expected a value but it was not there
                else:
                    errs.append(DataValidationError(f"missing attribute '{name}'.", object_path))
            except DataValidationError as e:
                errs.append(e)

        if len(errs) == 1:
            raise errs[0]
        if len(errs) > 1:
            raise AggregateDataValidationError(object_path, errs)
        return used_keys

    def _get_converted_value(self, obj: Any, key: str, source: TSource, object_path: str) -> Any:
        """Get a value of a field by invoking appropriate transformation function."""
        try:
            func = getattr(obj.__class__, f"_{key}")
            argc = len(inspect.signature(func).parameters)
            if argc == 1:
                # it is a static method
                return func(source)
            if argc == 2:
                # it is a instance method
                return func(_create_untouchable("obj"), source)
            raise RuntimeError("Transformation function has wrong number of arguments")
        except ValueError as e:
            msg = e.args[0] if len(e.args) > 0 and isinstance(e.args[0], str) else "Failed to validate value type"
            raise DataValidationError(msg, object_path) from e

    def object_constructor(self, obj: Any, source: Union["BaseSchema", Dict[Any, Any]], object_path: str) -> None:
        """
        Construct object. Delegated constructor for the NoRenameBaseSchema class.

        The reason this method is delegated to the mapper is due to renaming. Like this, we don't have to
        worry about a different BaseSchema class, when we want to have dynamically renamed fields.
        """
        # As this is a delegated constructor, we must ignore protected access warnings

        # sanity check
        if not isinstance(source, (BaseSchema, dict)):
            raise DataValidationError(f"expected dict-like object, found '{type(source)}'", object_path)

        # construct lower level schema first if configured to do so
        if obj._LAYER is not None:  # noqa: SLF001
            source = obj._LAYER(source, object_path=object_path)  # pylint: disable=not-callable  # noqa: SLF001

        # assign fields
        used_keys = self._assign_fields(obj, source, object_path)

        # check for unused keys in the source object
        if source and not isinstance(source, BaseSchema):
            unused = source.keys() - used_keys
            if len(unused) > 0:
                keys = ", ".join(f"'{u}'" for u in unused)
                raise DataValidationError(
                    f"unexpected extra key(s) {keys}",
                    object_path,
                )

        # validate the constructed value
        try:
            obj._validate()  # noqa: SLF001
        except ValueError as e:
            raise DataValidationError(e.args[0] if len(e.args) > 0 else "Validation error", object_path or "/") from e


class BaseSchema(Serializable):
    """
    Base class for modeling configuration schema.

    It somewhat resembles standard dataclasses with additional functionality:

    * type validation
    * data conversion

    To create an instance of this class, you have to provide source data in the form of dict-like object.
    Generally, raw dict or another `BaseSchema` instance. The provided data object is traversed, transformed
    and validated before assigned to the appropriate fields (attributes).

    Fields (attributes)
    ===================

    The fields (or attributes) of the class are defined the same way as in a dataclass by creating a class-level
    type-annotated fields. An example of that is:

    class A(BaseSchema):
        awesome_number: int

    If your `BaseSchema` instance has a field with type of a BaseSchema, its value is recursively created
    from the nested input data. This way, you can specify a complex tree of BaseSchema's and use the root
    BaseSchema to create instance of everything.

    Transformation
    ==============

    You can provide the BaseSchema class with a field and a function with the same name, but starting with
    underscore ('_'). For example, you could have field called `awesome_number` and function called
    `_awesome_number(self, source)`. The function takes one argument - the source data (optionally with self,
    but you are not supposed to touch that). It can read any data from the source object and return a value of
    an appropriate type, which will be assigned to the field `awesome_number`. If you want to report an error
    during validation, raise a `ValueError` exception.

    Using this, you can convert any input values into any type and field you want. To make the conversion easier
    to write, you could also specify a special class variable called `_LAYER` pointing to another
    BaseSchema class. This causes the source object to be first parsed as the specified additional layer of BaseSchema
    and after that used a source for this class. This therefore allows nesting of transformation functions.

    Validation
    ==========

    All assignments to fields during object construction are checked at runtime for proper types. This means,
    you are free to use an untrusted source object and turn it into a data structure, where you are sure what
    is what.

    You can also define a `_validate` method, which will be called once the whole data structure is built. You
    can validate the data in there and raise a `ValueError`, if they are invalid.

    Default values
    ==============

    If you create a field with a value, it will be used as a default value whenever the data in source object
    are not present. As a special case, default value for Optional type is None if not specified otherwise. You
    are not allowed to have a field with a default value and a transformation function at once.

    Example:
    =======

    See tests/utils/test_modelling.py for example usage.

    """

    _LAYER: Optional[Type["BaseSchema"]] = None
    _MAPPER: ObjectMapper = ObjectMapper()

    def __init__(self, source: TSource = None, object_path: str = "") -> None:  # pylint: disable=[super-init-not-called]
        # save source data (and drop information about nullness)
        source = source or {}
        self.__source: Union[Dict[str, Any], BaseSchema] = source

        # delegate the rest of the constructor
        self._MAPPER.object_constructor(self, source, object_path)

    def get_unparsed_data(self) -> Dict[str, Any]:
        if isinstance(self.__source, BaseSchema):
            return self.__source.get_unparsed_data()
        if isinstance(self.__source, Renamed):
            return self.__source.original()
        return self.__source

    def __getitem__(self, key: str) -> Any:
        if not hasattr(self, key):
            raise RuntimeError(f"Object '{self}' of type '{type(self)}' does not have field named '{key}'")
        return getattr(self, key)

    def __contains__(self, item: Any) -> bool:
        return hasattr(self, item)

    def _validate(self) -> None:
        """
        Additional validation procedure called after all field are assigned.

        Should throw a ValueError in case of failure.
        """

    def __eq__(self, o: object) -> bool:
        cls = self.__class__
        if not isinstance(o, cls):
            return False

        annot = get_annotations(cls)
        return all(getattr(self, name) == getattr(o, name) for name in annot)

    @classmethod
    def json_schema(
        cls: Type["BaseSchema"],
        schema_id: Optional[str] = None,
        title: Optional[str] = None,
        description: Optional[str] = None,
        include_schema_definition: bool = True,
    ) -> Dict[Any, Any]:
        if cls._LAYER is not None:
            return cls._LAYER.json_schema(
                schema_id=schema_id,
                title=title,
                description=description,
                include_schema_definition=include_schema_definition,
            )

        schema: Dict[Any, Any] = {}
        if include_schema_definition:
            schema["$schema"] = "https://json-schema.org/draft/2020-12/schema"
        if schema_id is not None:
            schema["$id"] = schema_id
        if title is not None:
            schema["title"] = title
        if description is not None:
            schema["description"] = description
        elif cls.__doc__ is not None:
            schema["description"] = _split_docstring(cls.__doc__)[0]
        schema["type"] = "object"
        schema["properties"] = _get_properties_schema(cls)

        return schema

    def to_dict(self) -> Dict[Any, Any]:
        res: Dict[Any, Any] = {}
        cls = self.__class__
        annot = get_annotations(cls)

        for name in annot:
            res[name] = Serializable.serialize(getattr(self, name))
        return res


class RenamingObjectMapper(ObjectMapper):
    """
    Same as object mapper, but it uses collection wrappers from the module `renamed` to perform dynamic field renaming.

    More specifically:
    - it renames all properties in (nested) objects
    - it does not rename keys in dictionaries
    """

    def _create_dict(self, tp: Type[Any], obj: Dict[Any, Any], object_path: str) -> Dict[Any, Any]:
        if isinstance(obj, Renamed):
            obj = obj.original()
        return super()._create_dict(tp, obj, object_path)

    def _create_base_schema_object(self, tp: Type[Any], obj: Any, object_path: str) -> "BaseSchema":
        if isinstance(obj, dict):
            obj = renamed(obj)
        return super()._create_base_schema_object(tp, obj, object_path)

    def object_constructor(self, obj: Any, source: Union["BaseSchema", Dict[Any, Any]], object_path: str) -> None:
        if isinstance(source, dict):
            source = renamed(source)
        return super().object_constructor(obj, source, object_path)


# export as a standalone functions for simplicity compatibility
is_obj_type_valid = ObjectMapper().is_obj_type_valid
map_object = ObjectMapper().map_object


class ConfigSchema(BaseSchema):
    """Same as BaseSchema, but maps with RenamingObjectMapper."""

    _MAPPER: ObjectMapper = RenamingObjectMapper()



# The 'typing.Pattern' is deprecated since python 3.8 and is removed in version 3.12.
# https://docs.python.org/3.9/library/typing.html#typing.Pattern
try:
    from typing import Pattern
except ImportError:
    from re import Pattern

__all__ = ["Pattern"]



from typing import Literal

from knot_resolver.utils.modeling import ConfigSchema

GlueCheckingEnum = Literal["normal", "strict", "permissive"]


class OptionsSchema(ConfigSchema):
    """
    Fine-tuning global parameters of DNS resolver operation.

    ---
    glue_checking: Glue records strictness checking level.
    minimize: Send minimum amount of information in recursive queries to enhance privacy.
    query_loopback: Permits queries to loopback addresses.
    reorder_rrset: Controls whether resource records within a RRSet are reordered each time it is served from the cache.
    query_case_randomization: Randomize Query Character Case.
    priming: Initializing DNS resolver cache with Priming Queries (RFC 8109)
    rebinding_protection: Protection against DNS Rebinding attack.
    refuse_no_rd: Queries without RD (recursion desired) bit set in query are answered with REFUSED.
    time_jump_detection: Detection of difference between local system time and expiration time bounds in DNSSEC signatures for '. NS' records.
    violators_workarounds: Workarounds for known DNS protocol violators.
    serve_stale: Allows using timed-out records in case DNS resolver is unable to contact upstream servers.
    """  # noqa: E501

    glue_checking: GlueCheckingEnum = "normal"
    minimize: bool = True
    query_loopback: bool = False
    reorder_rrset: bool = True
    query_case_randomization: bool = True
    priming: bool = True
    rebinding_protection: bool = False
    refuse_no_rd: bool = True
    time_jump_detection: bool = True
    violators_workarounds: bool = False
    serve_stale: bool = False



from abc import ABC, abstractmethod  # pylint: disable=[no-name-in-module]
from typing import Any, Dict, Type


class BaseTypeABC(ABC):
    @abstractmethod
    def __init__(self, source_value: Any, object_path: str = "/") -> None:
        pass

    @abstractmethod
    def __int__(self) -> int:
        raise NotImplementedError(f" return 'int()' value for {type(self).__name__} is not implemented.")

    @abstractmethod
    def __str__(self) -> str:
        raise NotImplementedError(f"return 'str()' value for {type(self).__name__} is not implemented.")

    @abstractmethod
    def serialize(self) -> Any:
        """
        Dump configuration to JSON-serializable object.

        Returns a JSON-serializable object from which the object
        can be recreated again using the constructor.

        It's not necessary to return the same structure that was given as an input. It only has
        to be the same semantically.
        """
        raise NotImplementedError(f"{type(self).__name__}'s' 'serialize()' not implemented.")


class BaseValueType(BaseTypeABC):
    """
    Subclasses of this class can be used as type annotations in 'DataParser'.

    When a value is being parsed from a serialized format (e.g. JSON/YAML),
    an object will be created by calling the constructor of the appropriate
    type on the field value. The only limitation is that the value MUST NOT be `None`.

    There is no validation done on the wrapped value. The only condition is that
    it can't be `None`. If you want to perform any validation during creation,
    raise a `ValueError` in case of errors.
    """

    @classmethod
    @abstractmethod
    def json_schema(cls: Type["BaseValueType"]) -> Dict[Any, Any]:
        raise NotImplementedError()



import asyncio
import logging
from typing import TYPE_CHECKING, Dict, List, Tuple

from .exceptions import KresSubprocessControllerError

if TYPE_CHECKING:
    from knot_resolver.controller.interface import KresID, Subprocess


logger = logging.getLogger(__name__)


_REGISTERED_WORKERS: "Dict[KresID, Subprocess]" = {}


def get_registered_workers_kresids() -> "List[KresID]":
    return list(_REGISTERED_WORKERS.keys())


async def command_single_registered_worker(cmd: str) -> "Tuple[KresID, object]":
    for sub in _REGISTERED_WORKERS.values():
        return sub.id, await sub.command(cmd)
    raise KresSubprocessControllerError(
        "Unable to execute the command. There is no kresd worker running to execute the command."
        "Try start/restart the resolver.",
    )


async def command_registered_workers(cmd: str) -> "Dict[KresID, object]":
    async def single_pair(sub: "Subprocess") -> "Tuple[KresID, object]":
        return sub.id, await sub.command(cmd)

    pairs = await asyncio.gather(*(single_pair(inst) for inst in _REGISTERED_WORKERS.values()))
    return dict(pairs)


def unregister_worker(subprocess: "Subprocess") -> None:
    """Unregister kresd worker "Subprocess" from the list."""
    del _REGISTERED_WORKERS[subprocess.id]


def register_worker(subprocess: "Subprocess") -> None:
    """Register kresd worker "Subprocess" on the list."""
    _REGISTERED_WORKERS[subprocess.id] = subprocess



# noqa: INP001
import argparse
import sys
from enum import Enum
from typing import List, Literal, Optional, Tuple, Type

from knot_resolver.client.command import COMP_NOSPACE, Command, CommandArgs, CompWords, comp_get_words, register_command
from knot_resolver.datamodel import KresConfig
from knot_resolver.utils.modeling.parsing import DataFormat, parse_json, try_to_parse
from knot_resolver.utils.requests import request


class Operations(Enum):
    SET = 0
    DELETE = 1
    GET = 2


def operation_to_method(operation: Operations) -> Literal["PUT", "GET", "DELETE"]:
    if operation == Operations.SET:
        return "PUT"
    if operation == Operations.DELETE:
        return "DELETE"
    return "GET"


@register_command
class ConfigCommand(Command):
    def __init__(self, namespace: argparse.Namespace) -> None:
        super().__init__(namespace)
        self.path: str = str(namespace.path) if hasattr(namespace, "path") else ""
        self.format: DataFormat = namespace.format if hasattr(namespace, "format") else DataFormat.JSON
        self.operation: Optional[Operations] = namespace.operation if hasattr(namespace, "operation") else None
        self.file: Optional[str] = namespace.file if hasattr(namespace, "file") else None

    @staticmethod
    def register_args_subparser(
        subparser: "argparse._SubParsersAction[argparse.ArgumentParser]",
    ) -> Tuple[argparse.ArgumentParser, "Type[Command]"]:
        config = subparser.add_parser("config", help="Performs operations on the running resolver's configuration.")
        path_help = "Optional, path (JSON pointer, RFC6901) to the configuration resources. "
        " By default, the entire configuration is selected."

        config_subparsers = config.add_subparsers(help="operation type")

        # GET operation
        get_op = config_subparsers.add_parser("get", help="Get current configuration from the resolver.")
        get_op.set_defaults(operation=Operations.GET, format=DataFormat.YAML)

        get_op.add_argument(
            "-p",
            "--path",
            help=path_help,
            action="store",
            type=str,
            default="",
        )
        get_op.add_argument(
            "file",
            help="Optional, path to the file where to save exported configuration data."
            " If not specified, data will be printed.",
            type=str,
            nargs="?",
        )

        get_formats = get_op.add_mutually_exclusive_group()
        get_formats.add_argument(
            "--json",
            help="Get configuration data in JSON format.",
            const=DataFormat.JSON,
            action="store_const",
            dest="format",
        )
        get_formats.add_argument(
            "--yaml",
            help="Get configuration data in YAML format, default.",
            const=DataFormat.YAML,
            action="store_const",
            dest="format",
        )

        # SET operation
        set_op = config_subparsers.add_parser("set", help="Set new configuration for the resolver.")
        set_op.set_defaults(operation=Operations.SET)

        set_op.add_argument(
            "-p",
            "--path",
            help=path_help,
            action="store",
            type=str,
            default="",
        )

        value_or_file = set_op.add_mutually_exclusive_group()
        value_or_file.add_argument(
            "file",
            help="Optional, path to file with new configuration.",
            type=str,
            nargs="?",
        )
        value_or_file.add_argument(
            "value",
            help="Optional, new configuration value.",
            type=str,
            nargs="?",
        )

        # DELETE operation
        delete_op = config_subparsers.add_parser(
            "delete", help="Delete given configuration property or list item at the given index."
        )
        delete_op.set_defaults(operation=Operations.DELETE)
        delete_op.add_argument(
            "-p",
            "--path",
            help=path_help,
            action="store",
            type=str,
            default="",
        )
        return config, ConfigCommand

    @staticmethod
    def completion(args: List[str], parser: argparse.ArgumentParser) -> CompWords:
        nargs = len(args)

        if nargs > 1 and args[-2] in ["-p", "--path"]:
            words: CompWords = {}
            words[COMP_NOSPACE] = None

            path = args[-1]
            path_nodes = path.split("/")

            prefix = ""
            properties = KresConfig.json_schema()["properties"]
            is_list = False
            for i, node in enumerate(path_nodes):
                # first node is empty string
                if i == 0:
                    continue

                if node in properties:
                    is_list = False
                    if "properties" in properties[node]:
                        properties = properties[node]["properties"]
                        prefix += f"/{node}"
                        continue
                    if "items" in properties[node]:
                        properties = properties[node]["items"]["properties"]
                        prefix += f"/{node}"
                        is_list = True
                        continue
                    del words[COMP_NOSPACE]
                    break
                if is_list and node.isnumeric():
                    prefix += f"/{node}"
                    continue

            for key in properties.keys():
                words[f"{prefix}/{key}"] = properties[key]["description"]

            return words

        return comp_get_words(args, parser)

    def run(self, args: CommandArgs) -> None:
        if not self.operation:
            args.subparser.print_help()
            sys.exit()

        new_config = None
        path = f"v1/config{self.path}"
        method = operation_to_method(self.operation)

        if self.operation == Operations.SET:
            if self.file:
                try:
                    with open(self.file, "r") as f:
                        new_config = f.read()
                except FileNotFoundError:
                    new_config = self.file
            else:
                # use STDIN also when file is not specified
                new_config = input("Type new configuration: ")

        body = DataFormat.JSON.dict_dump(try_to_parse(new_config)) if new_config else None
        response = request(args.socket, method, path, body)

        if response.status != 200:
            print(response, file=sys.stderr)
            sys.exit(1)

        if self.operation == Operations.GET and self.file:
            with open(self.file, "w") as f:
                f.write(self.format.dict_dump(parse_json(response.body), indent=4))
            print(f"saved to: {self.file}")
        elif response.body:
            print(self.format.dict_dump(parse_json(response.body), indent=4))



import logging
from threading import Timer
from typing import Dict, Optional
from urllib.parse import quote

from knot_resolver.controller.registered_workers import command_registered_workers
from knot_resolver.datamodel import KresConfig
from knot_resolver.utils import compat
from knot_resolver.utils.requests import SocketDesc, request

logger = logging.getLogger(__name__)

_triggers: Optional["Triggers"] = None


class Triggers:
    def __init__(self, config: KresConfig) -> None:
        self._config = config

        self._reload_force = False
        self._renew_force = False
        self._renew_timer: Optional[Timer] = None
        self._reload_timer: Optional[Timer] = None
        self._cmd_timers: Dict[str, Timer] = {}

        management = config.management
        socket = SocketDesc(
            f'http+unix://{quote(str(management.unix_socket), safe="")}/',
            'Key "/management/unix-socket" in validated configuration',
        )
        if management.interface:
            socket = SocketDesc(
                f"http://{management.interface.addr}:{management.interface.port}",
                'Key "/management/interface" in validated configuration',
            )
        self._socket = socket

    def trigger_cmd(self, cmd: str) -> None:
        def _cmd() -> None:
            if compat.asyncio.is_event_loop_running():
                compat.asyncio.create_task(command_registered_workers(cmd))
            else:
                compat.asyncio.run(command_registered_workers(cmd))
            logger.info(f"Sending '{cmd}' command to reload watched files has finished")

        # skipping if command was already triggered
        if cmd in self._cmd_timers and self._cmd_timers[cmd].is_alive():
            logger.info(f"Skipping sending '{cmd}' command, it was already triggered")
            return
        # start a 5sec timer
        logger.info(f"Delayed send of '{cmd}' command has started")
        self._cmd_timers[cmd] = Timer(5, _cmd)
        self._cmd_timers[cmd].start()

    def cancel_cmd(self, cmd: str) -> None:
        if cmd in self._cmd_timers:
            self._cmd_timers[cmd].cancel()

    def trigger_renew(self, force: bool = False) -> None:
        def _renew() -> None:
            response = request(self._socket, "POST", "renew/force" if force else "renew")
            if response.status != 200:
                logger.error(f"Failed to renew configuration: {response.body}")
            logger.info("Renewing configuration has finished")
            self._renew_force = False

        # do not trigger renew if reload is scheduled
        if self._reload_timer and self._reload_timer.is_alive() and self._reload_force >= force:
            logger.info("Skipping renewing configuration, reload was already triggered")
            return

        # skipping if reload was already triggered
        if self._renew_timer and self._renew_timer.is_alive():
            if self._renew_force >= force:
                logger.info("Skipping renewing configuration, it was already triggered")
                return
            self._renew_timer.cancel()
            self._renew_force = False

        logger.info("Delayed configuration renew has started")
        # start a 5sec timer
        self._renew_timer = Timer(5, _renew)
        self._renew_timer.start()
        self._renew_force = force

    def trigger_reload(self, force: bool = False) -> None:
        def _reload() -> None:
            response = request(self._socket, "POST", "reload/force" if force else "reload")
            if response.status != 200:
                logger.error(f"Failed to reload configuration: {response.body}")
            logger.info("Reloading configuration has finished")
            self._reload_force = False

        # cancel renew
        if self._renew_timer and self._renew_timer.is_alive() and force >= self._renew_force:
            self._renew_timer.cancel()
            self._renew_force = False

        # skipping if reload was already triggered
        if self._reload_timer and self._reload_timer.is_alive():
            if self._reload_force >= force:
                logger.info("Skipping reloading configuration, it was already triggered")
                return
            logger.info("Cancelling already scheduled configuration reload, force reload triggered")
            self._reload_timer.cancel()
            self._reload_force = False

        logger.info("Delayed configuration reload has started")
        # start a 5sec timer
        self._reload_timer = Timer(5, _reload)
        self._reload_timer.start()
        self._reload_force = force


def trigger_cmd(config: KresConfig, cmd: str) -> None:
    global _triggers
    if not _triggers:
        _triggers = Triggers(config)
    _triggers.trigger_cmd(cmd)


def cancel_cmd(cmd: str) -> None:
    global _triggers  # noqa: PLW0602
    if _triggers:
        _triggers.cancel_cmd(cmd)


def trigger_renew(config: KresConfig, force: bool = False) -> None:
    global _triggers
    if not _triggers:
        _triggers = Triggers(config)
    _triggers.trigger_renew(force)


def trigger_reload(config: KresConfig, force: bool = False) -> None:
    global _triggers
    if not _triggers:
        _triggers = Triggers(config)
    _triggers.trigger_reload(force)



import argparse
import importlib
import os
import sys

from knot_resolver.constants import VERSION

from .client import KRES_CLIENT_NAME, KresClient
from .command import install_commands_parsers


def auto_import_commands() -> None:
    prefix = f"{'.'.join(__name__.split('.')[:-1])}.commands."
    for module_name in os.listdir(os.path.dirname(__file__) + "/commands"):
        if module_name[-3:] != ".py":
            continue
        importlib.import_module(f"{prefix}{module_name[:-3]}")


def create_main_argument_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        KRES_CLIENT_NAME,
        description="Knot Resolver command-line utility that serves as a client for"
        " communicating with the Knot Resolver management API."
        " The utility also provides tools to work with the resolver's"
        " declarative configuration (validate, convert, ...).",
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=VERSION,
        help="Get version",
    )
    # parser.add_argument(
    #     "-i",
    #     "--interactive",
    #     action="store_true",
    #     help="Use the utility in interactive mode.",
    #     default=False,
    #     required=False,
    # )
    config_or_socket = parser.add_mutually_exclusive_group()
    config_or_socket.add_argument(
        "-s",
        "--socket",
        action="store",
        type=str,
        help="Optional, path to the resolver's management API, unix-domain socket, or network interface."
        " Cannot be used together with '--config'.",
        default=[],
        nargs=1,
        required=False,
    )
    config_or_socket.add_argument(
        "-c",
        "--config",
        action="store",
        type=str,
        help="Optional, path to the resolver's declarative configuration to retrieve the management API configuration."
        " Cannot be used together with '--socket'.",
        default=[],
        nargs=1,
        required=False,
    )
    return parser


def main() -> None:
    auto_import_commands()
    parser = create_main_argument_parser()
    install_commands_parsers(parser)

    # TODO: This is broken with unpatched versions of poethepoet, because they drop the `--` pseudo-argument.
    # Patch submitted at <https://github.com/nat-n/poethepoet/pull/163>.
    try:
        pa_index = sys.argv.index("--", 1)
        argv_to_parse = sys.argv[1:pa_index]
        argv_extra = sys.argv[(pa_index + 1) :]
    except ValueError:
        argv_to_parse = sys.argv[1:]
        argv_extra = None

    namespace = parser.parse_args(argv_to_parse)
    if hasattr(namespace, "extra"):
        raise TypeError("'extra' is already an attribute - this is disallowed for commands")
    namespace.extra = argv_extra

    client = KresClient(namespace, parser)
    client.execute()

    # if namespace.interactive or len(vars(namespace)) == 2:
    #     client.interactive()
    # else:
    #     client.execute()



import logging
from os import getppid, kill  # pylint: disable=[no-name-in-module]
from pathlib import Path
from typing import Any, Dict, Iterable, NoReturn, Optional, Union, cast
from xmlrpc.client import Fault, ServerProxy

import supervisor.xmlrpc  # type: ignore[import]

from knot_resolver.controller.exceptions import KresSubprocessControllerError, KresSubprocessControllerExec
from knot_resolver.controller.interface import (
    KresID,
    Subprocess,
    SubprocessController,
    SubprocessStatus,
    SubprocessType,
)
from knot_resolver.controller.supervisord.config_file import SupervisordKresID, write_config_file
from knot_resolver.datamodel.config_schema import KresConfig, workers_max_count
from knot_resolver.manager.constants import supervisord_config_file, supervisord_pid_file, supervisord_sock_file
from knot_resolver.utils import which
from knot_resolver.utils.async_utils import call, readfile
from knot_resolver.utils.compat.asyncio import async_in_a_thread

logger = logging.getLogger(__name__)


async def _start_supervisord(config: KresConfig) -> None:
    logger.debug("Writing supervisord config")
    await write_config_file(config)
    logger.debug("Starting supervisord")
    res = await call(["supervisord", "--configuration", str(supervisord_config_file(config).absolute())])
    if res != 0:
        raise KresSubprocessControllerError(f"Supervisord exited with exit code {res}")


async def _exec_supervisord(config: KresConfig) -> NoReturn:
    logger.debug("Writing supervisord config")
    await write_config_file(config)
    logger.debug("Execing supervisord")
    raise KresSubprocessControllerExec(
        [
            str(which.which("supervisord")),
            "supervisord",
            "--configuration",
            str(supervisord_config_file(config).absolute()),
        ]
    )


async def _reload_supervisord(config: KresConfig) -> None:
    await write_config_file(config)
    try:
        supervisord = _create_supervisord_proxy(config)
        supervisord.reloadConfig()
    except Fault as e:
        raise KresSubprocessControllerError(f"supervisord reload failed: {e}") from e


@async_in_a_thread
def _stop_supervisord(config: KresConfig) -> None:
    supervisord = _create_supervisord_proxy(config)
    # pid = supervisord.getPID()
    try:
        # we might be trying to shut down supervisord at a moment, when it's waiting
        # for us to stop. Therefore, this shutdown request for supervisord might
        # die and it's not a problem.
        supervisord.shutdown()
    except Fault as e:
        if e.faultCode == 6 and e.faultString == "SHUTDOWN_STATE":
            # supervisord is already stopping, so it's fine
            pass
        else:
            # something wrong happened, let's be loud about it
            raise

    # It is always better to clean up.
    # This way, we can be sure that we are starting with a newly generated configuration.
    supervisord_config_file(config).unlink()


async def _is_supervisord_available() -> bool:
    # yes, it is! The code in this file wouldn't be running without it due to imports :)

    # so let's just check that we can find supervisord and supervisorctl binaries
    try:
        which.which("supervisord")
        which.which("supervisorctl")
    except RuntimeError:
        logger.error("Failed to find supervisord or supervisorctl executables in $PATH")
        return False

    return True


async def _get_supervisord_pid(config: KresConfig) -> Optional[int]:
    if not Path(supervisord_pid_file(config)).exists():
        return None

    return int(await readfile(supervisord_pid_file(config)))


def _is_process_runinng(pid: int) -> bool:
    try:
        # kill with signal 0 is a safe way to test that a process exists
        kill(pid, 0)
    except ProcessLookupError:
        return False
    else:
        return True


async def _is_supervisord_running(config: KresConfig) -> bool:
    pid = await _get_supervisord_pid(config)
    if pid is None:
        return False
    if not _is_process_runinng(pid) or getppid() != pid:
        supervisord_pid_file(config).unlink()
        return False
    return True


def _create_proxy(config: KresConfig) -> ServerProxy:
    return ServerProxy(
        "http://127.0.0.1",
        transport=supervisor.xmlrpc.SupervisorTransport(
            None, None, serverurl="unix://" + str(supervisord_sock_file(config))
        ),
    )


def _create_supervisord_proxy(config: KresConfig) -> Any:
    proxy = _create_proxy(config)
    return getattr(proxy, "supervisor")


def _create_fast_proxy(config: KresConfig) -> Any:
    proxy = _create_proxy(config)
    return getattr(proxy, "fast")


def _convert_subprocess_status(proc: Any) -> SubprocessStatus:
    conversion_tbl = {
        # "STOPPED": None,  # filtered out elsewhere
        "STARTING": SubprocessStatus.RUNNING,
        "RUNNING": SubprocessStatus.RUNNING,
        "BACKOFF": SubprocessStatus.RUNNING,
        "STOPPING": SubprocessStatus.RUNNING,
        "EXITED": SubprocessStatus.EXITED,
        "FATAL": SubprocessStatus.FATAL,
        "UNKNOWN": SubprocessStatus.UNKNOWN,
    }

    if proc["statename"] in conversion_tbl:
        status = conversion_tbl[proc["statename"]]
    else:
        logger.warning(f"Unknown supervisord process state {proc['statename']}")
        status = SubprocessStatus.UNKNOWN
    return status


def _list_running_subprocesses(config: KresConfig) -> Dict[SupervisordKresID, SubprocessStatus]:
    try:
        supervisord = _create_supervisord_proxy(config)
        processes: Any = supervisord.getAllProcessInfo()
    except Fault as e:
        raise KresSubprocessControllerError(f"failed to get info from all running processes: {e}") from e

    # there will be a manager process as well, but we don't want to report anything on ourselves
    processes = [pr for pr in processes if pr["name"] != "manager"]

    # convert all the names
    return {
        SupervisordKresID.from_string(f"{pr['group']}:{pr['name']}"): _convert_subprocess_status(pr)
        for pr in processes
        if pr["statename"] != "STOPPED"
    }


class SupervisordSubprocess(Subprocess):
    def __init__(
        self,
        config: KresConfig,
        controller: "SupervisordSubprocessController",
        base_id: Union[SubprocessType, SupervisordKresID],
    ) -> None:
        if isinstance(base_id, SubprocessType):
            super().__init__(config, SupervisordKresID.alloc(base_id))
        else:
            super().__init__(config, base_id)
        self._controller: "SupervisordSubprocessController" = controller

    @property
    def name(self) -> str:
        return str(self.id)

    def status(self) -> SubprocessStatus:
        try:
            supervisord = _create_supervisord_proxy(self._config)
            status = supervisord.getProcessInfo(self.name)
        except Fault as e:
            raise KresSubprocessControllerError(f"failed to get status from '{self.id}' process: {e}") from e
        return _convert_subprocess_status(status)

    @async_in_a_thread
    def _start(self) -> None:
        # +1 for canary process (same as in config_file.py)
        assert int(self.id) <= int(workers_max_count()) + 1, "trying to spawn more than allowed limit of workers"
        try:
            supervisord = _create_fast_proxy(self._config)
            supervisord.startProcess(self.name)
        except Fault as e:
            raise KresSubprocessControllerError(f"failed to start '{self.id}'") from e

    @async_in_a_thread
    def _stop(self) -> None:
        supervisord = _create_supervisord_proxy(self._config)
        supervisord.stopProcess(self.name)

    @async_in_a_thread
    def _restart(self) -> None:
        supervisord = _create_supervisord_proxy(self._config)
        supervisord.stopProcess(self.name)
        fast = _create_fast_proxy(self._config)
        fast.startProcess(self.name)

    @async_in_a_thread
    def get_pid(self) -> int:
        if self._pid is None:
            supervisord = _create_supervisord_proxy(self._config)
            info = supervisord.getProcessInfo(self.name)
            self._pid = info["pid"]
        return self._pid

    def get_used_config(self) -> KresConfig:
        return self._config


class SupervisordSubprocessController(SubprocessController):
    def __init__(self) -> None:  # pylint: disable=super-init-not-called
        self._controller_config: Optional[KresConfig] = None

    def __str__(self) -> str:
        return "supervisord"

    async def is_controller_available(self, config: KresConfig) -> bool:
        res = await _is_supervisord_available()
        if not res:
            logger.info("Failed to find usable supervisord.")

        logger.debug("Detection - supervisord controller is available for use")
        return res

    async def get_all_running_instances(self) -> Iterable[Subprocess]:
        assert self._controller_config is not None

        if await _is_supervisord_running(self._controller_config):
            states = _list_running_subprocesses(self._controller_config)
            return [
                SupervisordSubprocess(self._controller_config, self, id_)
                for id_ in states
                if states[id_] == SubprocessStatus.RUNNING
            ]
        return []

    async def initialize_controller(self, config: KresConfig) -> None:
        self._controller_config = config

        if not await _is_supervisord_running(config):
            logger.info(
                "We want supervisord to restart us when needed, we will therefore exec() it and let it start us again."
            )
            await _exec_supervisord(config)
        else:
            logger.info("Supervisord is already running, we will just update its config...")
            await _reload_supervisord(config)

    async def shutdown_controller(self) -> None:
        assert self._controller_config is not None
        await _stop_supervisord(self._controller_config)

    async def create_subprocess(self, subprocess_config: KresConfig, subprocess_type: SubprocessType) -> Subprocess:
        return SupervisordSubprocess(subprocess_config, self, subprocess_type)

    @async_in_a_thread
    def get_subprocess_status(self) -> Dict[KresID, SubprocessStatus]:
        assert self._controller_config is not None
        return cast(Dict[KresID, SubprocessStatus], _list_running_subprocesses(self._controller_config))



import json
from enum import Enum, auto
from typing import Any, Dict, List, Optional, Tuple, Union

import yaml
from yaml.constructor import ConstructorError
from yaml.nodes import MappingNode

from .exceptions import DataParsingError, DataValidationError
from .renaming import Renamed, renamed


# custom hook for 'json.loads()' to detect duplicate keys in data
# source: https://stackoverflow.com/q/14902299/12858520
def _json_raise_duplicates(pairs: List[Tuple[Any, Any]]) -> Optional[Any]:
    dict_out: Dict[Any, Any] = {}
    for key, val in pairs:
        if key in dict_out:
            raise DataParsingError(f"Duplicate attribute key detected: {key}")
        dict_out[key] = val
    return dict_out


# custom loader for 'yaml.load()' to detect duplicate keys in data
# source: https://gist.github.com/pypt/94d747fe5180851196eb
class _RaiseDuplicatesLoader(yaml.SafeLoader):
    def construct_mapping(self, node: Union[MappingNode, Any], deep: bool = False) -> Dict[Any, Any]:
        if not isinstance(node, MappingNode):
            raise ConstructorError(None, None, f"expected a mapping node, but found {node.id}", node.start_mark)
        mapping: Dict[Any, Any] = {}
        for key_node, value_node in node.value:
            key = self.construct_object(key_node, deep=deep)
            # we need to check, that the key object can be used in a hash table
            try:
                _ = hash(key)
            except TypeError as exc:
                raise ConstructorError(
                    "while constructing a mapping",
                    node.start_mark,
                    f"found unacceptable key ({exc})",
                    key_node.start_mark,
                ) from exc

            # check for duplicate keys
            if key in mapping:
                raise DataParsingError(f"duplicate key detected: {key_node.start_mark}")
            value = self.construct_object(value_node, deep=deep)
            mapping[key] = value
        return mapping


class DataFormat(Enum):
    YAML = auto()
    JSON = auto()

    def parse_to_dict(self, text: str) -> Any:
        if self is DataFormat.YAML:
            # RaiseDuplicatesLoader extends yaml.SafeLoader, so this should be safe
            # https://python.land/data-processing/python-yaml#PyYAML_safe_load_vs_load
            return renamed(yaml.load(text, Loader=_RaiseDuplicatesLoader))  # noqa: S506
        if self is DataFormat.JSON:
            return renamed(json.loads(text, object_pairs_hook=_json_raise_duplicates))
        raise NotImplementedError(f"Parsing of format '{self}' is not implemented")

    def dict_dump(self, data: Union[Dict[str, Any], Renamed], indent: Optional[int] = None) -> str:
        if isinstance(data, Renamed):
            data = data.original()

        if self is DataFormat.YAML:
            return yaml.safe_dump(data, indent=indent)
        if self is DataFormat.JSON:
            return json.dumps(data, indent=indent)
        raise NotImplementedError(f"Exporting to '{self}' format is not implemented")


def parse_yaml(data: str) -> Any:
    return DataFormat.YAML.parse_to_dict(data)


def parse_json(data: str) -> Any:
    return DataFormat.JSON.parse_to_dict(data)


def try_to_parse(data: str) -> Any:
    """Attempt to parse the data as a JSON or YAML string."""
    try:
        return parse_json(data)
    except json.JSONDecodeError as je:
        try:
            return parse_yaml(data)
        except yaml.YAMLError as ye:
            # We do not raise-from here because there are two possible causes
            # and we may not know which one is the actual one.
            raise DataParsingError(  # pylint: disable=raise-missing-from
                f"failed to parse data, JSON: {je}, YAML: {ye}"
            ) from ye


def data_combine(data: Dict[Any, Any], additional_data: Dict[Any, Any], object_path: str = "") -> Dict[Any, Any]:
    """Combine dictionaries data."""
    for key in additional_data:
        if key in data:
            # if both are dictionaries we can try to combine them deeper
            if isinstance(data[key], (Dict, dict)) and isinstance(additional_data[key], (Dict, dict)):
                data[key] = data_combine(data[key], additional_data[key], f"{object_path}/{key}").copy()
                continue
            # otherwise we cannot combine them
            raise DataValidationError(f"duplicity key '{key}' with value in data", object_path)
        val = additional_data[key]
        data[key] = val.copy() if hasattr(val, "copy") else val
    return data



# noqa: INP001
import argparse
import copy
import sys
from typing import Any, Dict, List, Optional, Tuple, Type

from knot_resolver.client.command import Command, CommandArgs, CompWords, comp_get_words, register_command
from knot_resolver.constants import VERSION
from knot_resolver.utils.modeling.exceptions import DataParsingError
from knot_resolver.utils.modeling.parsing import DataFormat, try_to_parse


def _remove(config: Dict[str, Any], path: str) -> Optional[Any]:
    keys = path.split("/")
    last = keys[-1]

    current = config
    for key in keys[1:-1]:
        if key in current:
            current = current[key]
        else:
            return None
    if isinstance(current, dict) and last in current:
        val = copy.copy(current[last])
        del current[last]
        print(f"removed {path}")
        return val
    return None


def _add(config: Dict[str, Any], path: str, val: Any, rewrite: bool = False) -> None:
    keys = path.split("/")
    last = keys[-1]

    current = config
    for key in keys[1:-1]:
        if key not in current or key in current and not isinstance(current[key], dict):
            current[key] = {}
        current = current[key]

    if rewrite or last not in current:
        current[last] = val
        print(f"added {path}")


def _rename(config: Dict[str, Any], path: str, new_path: str) -> None:
    val: Optional[Any] = _remove(config, path)
    if val:
        _add(config, new_path, val)


@register_command
class MigrateCommand(Command):
    def __init__(self, namespace: argparse.Namespace) -> None:
        super().__init__(namespace)
        self.input_file: str = namespace.input_file
        self.output_file: Optional[str] = namespace.output_file
        self.output_format: DataFormat = namespace.output_format

    @staticmethod
    def register_args_subparser(
        subparser: "argparse._SubParsersAction[argparse.ArgumentParser]",
    ) -> Tuple[argparse.ArgumentParser, "Type[Command]"]:
        migrate = subparser.add_parser("migrate", help="Migrates JSON or YAML configuration to the newer version.")

        migrate.set_defaults(output_format=DataFormat.YAML)
        output_formats = migrate.add_mutually_exclusive_group()
        output_formats.add_argument(
            "--json",
            help="Get migrated configuration data in JSON format.",
            const=DataFormat.JSON,
            action="store_const",
            dest="output_format",
        )
        output_formats.add_argument(
            "--yaml",
            help="Get migrated configuration data in YAML format, default.",
            const=DataFormat.YAML,
            action="store_const",
            dest="output_format",
        )

        migrate.add_argument(
            "input_file",
            type=str,
            help="File with configuration in YAML or JSON format.",
        )
        migrate.add_argument(
            "output_file",
            type=str,
            nargs="?",
            help="Optional, output file for migrated configuration in desired output format."
            " If not specified, migrated configuration is printed.",
            default=None,
        )
        return migrate, MigrateCommand

    @staticmethod
    def completion(args: List[str], parser: argparse.ArgumentParser) -> CompWords:
        return comp_get_words(args, parser)

    def run(self, args: CommandArgs) -> None:  # noqa: C901, PLR0912, PLR0915
        with open(self.input_file, "r") as f:
            data = f.read()

        try:
            parsed = try_to_parse(data)
        except DataParsingError as e:
            print(e, file=sys.stderr)
            sys.exit(1)

        new = parsed.copy()

        # REMOVE
        _remove(new, "/dnssec/refresh-time")
        _remove(new, "/dnssec/hold-down-time")
        _remove(new, "/dnssec/time-skew-detection")
        _remove(new, "/dnssec/keep-removed")
        _remove(new, "/local-data/root-fallback-addresses")
        _remove(new, "/local-data/root-fallback-addresses-files")
        _remove(new, "/logging/debugging")
        _remove(new, "/max-workers")
        _remove(new, "/network/tls/auto-discovery")
        _remove(new, "/webmgmt")

        # RENAME/MOVE
        cache_key = "cache"
        if cache_key in new:
            gc_key = "garbage-collector"
            if gc_key in new[cache_key]:
                gc = new[cache_key][gc_key]
                if gc is False:
                    _add(new, "/cache/garbage-collector/enable", False)
                else:
                    _add(new, "/cache/garbage-collector/enable", True)
            prefetch_key = "prefetch"
            if prefetch_key in new[cache_key]:
                prediction_key = "prediction"
                if prediction_key in new[cache_key][prefetch_key]:
                    prediction = new[cache_key][prefetch_key][prediction_key]
                    if prediction is None:
                        _add(new, "/cache/prefetch/prediction/enable", False)
                    else:
                        _add(new, "/cache/prefetch/prediction/enable", True)
        _rename(new, "/defer/enabled", "/defer/enable")
        dns64_key = "dns64"
        if dns64_key in new:
            if new[dns64_key] is False:
                _add(new, "/dns64/enable", False, rewrite=True)
            else:
                _add(new, "/dns64/enable", True, rewrite=True)
        _rename(new, "/dns64/rev-ttl", "/dns64/reverse-ttl")
        dnssec_key = "dnssec"
        if dnssec_key in new:
            if new[dnssec_key] is False:
                _add(new, "/dnssec/enable", False, rewrite=True)
            else:
                # by default the DNSSEC is enabled
                pass
        _rename(new, "/dnssec/trust-anchor-sentinel", "/dnssec/sentinel")
        _rename(new, "/dnssec/trust-anchor-signal-query", "/dnssec/signal-query")
        logging_key = "logging"
        if logging_key in new:
            dnstap_key = "dnstap"
            if dnstap_key in new[logging_key]:
                dnstap = new[logging_key][dnstap_key]
                if dnstap is None:
                    _add(new, "/logging/dnstap/enable", False)
                else:
                    _add(new, "/logging/dnstap/enable", True)

        _rename(new, "/logging/dnssec-bogus", "/dnssec/log-bogus")
        _rename(new, "/monitoring/enabled", "/monitoring/metrics")
        monitoring_key = "monitoring"
        if monitoring_key in new:
            graphite_key = "graphite"
            if graphite_key in new[monitoring_key]:
                graphite = new[monitoring_key][graphite_key]
                if graphite is False:
                    _add(new, "/monitoring/graphite/enable", False)
                else:
                    _add(new, "/monitoring/graphite/enable", True)
        network_key = "network"
        if network_key in new:
            proxy_protocol_key = "proxy-protocol"
            if proxy_protocol_key in new[network_key]:
                proxy_protocol = new[network_key][proxy_protocol_key]
                if proxy_protocol is None:
                    _add(new, "/network/proxy-protocol/enable", False)
                else:
                    _add(new, "/network/proxy-protocol/enable", True)
        _rename(new, "/network/tls/files-watchdog", "/network/tls/watchdog")
        rate_limiting_key = "rate-limiting"
        if rate_limiting_key in new:
            _add(new, "/rate-limiting/enable", True)

        # remove empty dicts
        new = {k: v for k, v in new.items() if v}

        dumped = self.output_format.dict_dump(new)
        if self.output_file:
            with open(self.output_file, "w") as f:
                f.write(dumped)
        else:
            print(f"\nNew migrated configuration (v{VERSION}):")
            print("---")
            print(dumped)



import os
import sys

from jinja2 import Environment, FileSystemLoader, StrictUndefined, Template


def _get_templates_dir() -> str:
    module = sys.modules["knot_resolver.datamodel"].__file__
    if module:
        templates_dir = os.path.join(os.path.dirname(module), "templates")
        if os.path.isdir(templates_dir):
            return templates_dir
        raise NotADirectoryError(f"the templates dir '{templates_dir}' is not a directory or does not exist")
    raise OSError("package 'knot_resolver.datamodel' cannot be located or loaded")


_TEMPLATES_DIR = _get_templates_dir()


def _import_kresd_config_template() -> Template:
    path = os.path.join(_TEMPLATES_DIR, "kresd.lua.j2")
    with open(path, "r", encoding="UTF-8") as file:
        template = file.read()
    return template_from_str(template)


def _import_policy_loader_config_template() -> Template:
    path = os.path.join(_TEMPLATES_DIR, "policy-loader.lua.j2")
    with open(path, "r", encoding="UTF-8") as file:
        template = file.read()
    return template_from_str(template)


def template_from_str(template: str) -> Template:
    ldr = FileSystemLoader(_TEMPLATES_DIR)
    env = Environment(trim_blocks=True, lstrip_blocks=True, loader=ldr, undefined=StrictUndefined)
    return env.from_string(template)


KRESD_CONFIG_TEMPLATE = _import_kresd_config_template()


POLICY_LOADER_CONFIG_TEMPLATE = _import_policy_loader_config_template()



import asyncio
import itertools
import json
import logging
import struct
import sys
from abc import ABC, abstractmethod  # pylint: disable=no-name-in-module
from enum import Enum, auto
from pathlib import Path
from typing import Dict, Iterable, Optional, Type, TypeVar
from weakref import WeakValueDictionary

from knot_resolver.controller.exceptions import KresSubprocessControllerError
from knot_resolver.controller.registered_workers import register_worker, unregister_worker
from knot_resolver.datamodel.config_schema import KresConfig
from knot_resolver.manager.constants import kresd_config_file, policy_loader_config_file

logger = logging.getLogger(__name__)


class SubprocessType(Enum):
    KRESD = auto()
    POLICY_LOADER = auto()
    GC = auto()


class SubprocessStatus(Enum):
    RUNNING = auto()
    FATAL = auto()
    EXITED = auto()
    UNKNOWN = auto()


T = TypeVar("T", bound="KresID")


class KresID:
    """ID object used for identifying subprocesses."""

    _used: "Dict[SubprocessType, WeakValueDictionary[int, KresID]]" = {k: WeakValueDictionary() for k in SubprocessType}

    @classmethod
    def alloc(cls: Type[T], typ: SubprocessType) -> T:
        # find free ID closest to zero
        for i in itertools.count(start=0, step=1):
            if i not in cls._used[typ]:
                return cls.new(typ, i)

        raise RuntimeError("Reached an end of an infinite loop. How?")

    @classmethod
    def new(cls: "Type[T]", typ: SubprocessType, n: int) -> "T":
        if n in cls._used[typ]:
            # Ignoring typing here, because I can't find a way how to make the _used dict
            # typed based on subclass. I am not even sure that it's different between subclasses,
            # it's probably still the same dict. But we don't really care about it
            return cls._used[typ][n]  # type: ignore[return-value]
        val = cls(typ, n, _i_know_what_i_am_doing=True)
        cls._used[typ][n] = val
        return val

    def __init__(self, typ: SubprocessType, n: int, _i_know_what_i_am_doing: bool = False) -> None:
        if not _i_know_what_i_am_doing:
            raise RuntimeError("Don't do this. You seem to have no idea what it does")

        self._id = n
        self._type = typ

    @property
    def subprocess_type(self) -> SubprocessType:
        return self._type

    def __repr__(self) -> str:
        return f"KresID({self})"

    def __hash__(self) -> int:
        return self._id

    def __eq__(self, o: object) -> bool:
        if isinstance(o, KresID):
            return self._type == o._type and self._id == o._id
        return False

    def __str__(self) -> str:
        """Return string representation of the ID usable directly in the underlying service supervisor."""
        raise NotImplementedError()

    @staticmethod
    def from_string(val: str) -> "KresID":
        """Inverse of __str__."""
        raise NotImplementedError()

    def __int__(self) -> int:
        return self._id


class Subprocess(ABC):
    """One SubprocessInstance corresponds to one manager's subprocess."""

    def __init__(self, config: KresConfig, kresid: KresID) -> None:
        self._id = kresid
        self._config = config
        self._registered_worker: bool = False
        self._pid: Optional[int] = None

        self._config_file: Optional[Path] = None
        if self.type is SubprocessType.KRESD:
            self._config_file = kresd_config_file(self._config, self.id)
        elif self.type is SubprocessType.POLICY_LOADER:
            self._config_file = policy_loader_config_file(self._config)

    def _render_lua(self) -> Optional[str]:
        if self.type is SubprocessType.KRESD:
            return self._config.render_kresd_lua()
        if self.type is SubprocessType.POLICY_LOADER:
            return self._config.render_policy_loader_lua()
        return None

    def _write_config(self) -> None:
        config_lua = self._render_lua()
        if config_lua and self._config_file:
            with open(self._config_file, "w", encoding="utf8") as file:
                file.write(config_lua)

    def _unlink_config(self) -> None:
        if self._config_file:
            self._config_file.unlink(missing_ok=True)

    async def start(self, new_config: Optional[KresConfig] = None) -> None:
        if new_config:
            self._config = new_config
        self._write_config()

        try:
            await self._start()
            if self.type is SubprocessType.KRESD:
                register_worker(self)
                self._registered_worker = True
        except KresSubprocessControllerError:
            self._unlink_config()
            raise

    async def apply_new_config(self, new_config: KresConfig) -> None:
        self._config = new_config

        # update config file
        logger.debug(f"Writing config file for {self.id}")
        self._write_config()

        # update runtime status
        logger.debug(f"Restarting {self.id}")
        await self._restart()

    async def stop(self) -> None:
        if self._registered_worker:
            unregister_worker(self)
        await self._stop()
        await self.cleanup()

    async def cleanup(self) -> None:
        """
        Remove temporary files and all traces of this instance running.

        It is NOT SAFE to call this while the kresd is running,
        because it will break automatic restarts (at the very least).
        """
        self._unlink_config()

    def __eq__(self, o: object) -> bool:
        return isinstance(o, type(self)) and o.type == self.type and o.id == self.id

    def __hash__(self) -> int:
        return hash(type(self)) ^ hash(self.type) ^ hash(self.id)

    @abstractmethod
    async def _start(self) -> None:
        pass

    @abstractmethod
    async def _stop(self) -> None:
        pass

    @abstractmethod
    async def _restart(self) -> None:
        pass

    @abstractmethod
    async def get_pid(self) -> int:
        pass

    @abstractmethod
    def status(self) -> SubprocessStatus:
        pass

    @property
    def type(self) -> SubprocessType:
        return self.id.subprocess_type

    @property
    def id(self) -> KresID:
        return self._id

    async def command(self, cmd: str) -> object:
        if not self._registered_worker:
            raise RuntimeError("the command cannot be sent to a process other than the kresd worker")

        reader: asyncio.StreamReader
        writer: Optional[asyncio.StreamWriter] = None

        try:
            reader, writer = await asyncio.open_unix_connection(f"./control/{int(self.id)}")

            # drop prompt
            _ = await reader.read(2)

            # switch to JSON mode
            writer.write("__json\n".encode("utf8"))

            # write command
            writer.write(cmd.encode("utf8"))
            writer.write(b"\n")
            await writer.drain()

            # read result
            (msg_len,) = struct.unpack(">I", await reader.read(4))
            result_bytes = await reader.readexactly(msg_len)

            try:
                return json.loads(result_bytes.decode("utf8"))
            except json.JSONDecodeError:
                return result_bytes.decode("utf8")

        finally:
            if writer is not None:
                writer.close()

                # proper closing of the socket is only implemented in later versions of python
                if sys.version_info >= (3, 7):
                    await writer.wait_closed()


class SubprocessController(ABC):
    """
    The common Subprocess Controller interface.

    This is what KresManager requires and what has to be implemented by all controllers.
    """

    @abstractmethod
    async def is_controller_available(self, config: KresConfig) -> bool:
        """Return bool, whether the controller is available with the given config."""

    @abstractmethod
    async def initialize_controller(self, config: KresConfig) -> None:
        """
        Initialize the Subprocess Controller.

        Should be called when we want to really start using the controller with a specific configuration.
        """

    @abstractmethod
    async def get_all_running_instances(self) -> Iterable[Subprocess]:
        """Must NOT be called before initialize_controller()."""

    @abstractmethod
    async def shutdown_controller(self) -> None:
        """
        Shutting the Process Cntroller.

        Allows us to stop the service manager process or simply cleanup,
        so that we don't reuse the same resources in a new run.

        Must NOT be called before initialize_controller()
        """

    @abstractmethod
    async def create_subprocess(self, subprocess_config: KresConfig, subprocess_type: SubprocessType) -> Subprocess:
        """
        Return a Subprocess object which can be operated on.

        The subprocess is not started or in any way active after this call.
        That has to be performaed manually using the returned object itself.

        Must NOT be called before initialize_controller()
        """

    @abstractmethod
    async def get_subprocess_status(self) -> Dict[KresID, SubprocessStatus]:
        """
        Get a status of running subprocesses as seen by the controller.

        This method  actively polls for information.

        Must NOT be called before initialize_controller()
        """



# noqa: INP001
import argparse
from enum import Enum
from typing import List, Tuple, Type

from knot_resolver.client.command import (
    Command,
    CommandArgs,
    CompWords,
    comp_get_words,
    register_command,
)


class Shells(Enum):
    BASH = 0
    FISH = 1


@register_command
class CompletionCommand(Command):
    def __init__(self, namespace: argparse.Namespace) -> None:
        super().__init__(namespace)
        self.shell: Shells = namespace.shell
        self.args: List[str] = namespace.args
        if namespace.extra is not None:
            self.args.append("--")

    @staticmethod
    def register_args_subparser(
        subparser: "argparse._SubParsersAction[argparse.ArgumentParser]",
    ) -> Tuple[argparse.ArgumentParser, "Type[Command]"]:
        completion = subparser.add_parser(
            "completion",
            help="commands auto-completion",
        )

        shells_dest = "shell"
        shells = completion.add_mutually_exclusive_group()
        shells.add_argument("--bash", action="store_const", dest=shells_dest, const=Shells.BASH, default=Shells.BASH)
        shells.add_argument("--fish", action="store_const", dest=shells_dest, const=Shells.FISH)

        completion.add_argument("--args", help="arguments to complete", nargs=argparse.REMAINDER, default=[])

        return completion, CompletionCommand

    @staticmethod
    def completion(args: List[str], parser: argparse.ArgumentParser) -> CompWords:
        return comp_get_words(args, parser)

    def run(self, args: CommandArgs) -> None:
        words: CompWords = {}

        parser = args.parser
        if parser:
            words = comp_get_words(self.args, args.parser)

        # print completion words
        # based on required bash/fish shell format
        if self.shell == Shells.BASH:
            print(" ".join(words))
        elif self.shell == Shells.FISH:
            # TODO: FISH completion implementation
            pass
        else:
            raise ValueError(f"unexpected value of {Shells}: {self.shell}")
